@@ -216,6 +216,8 @@ static void PianoXmlParseUserinfoCb (const char *key, const ezxml_t value,
 		user->webAuthToken = strdup (valueStr);
 	} else if (strcmp ("authToken", key) == 0) {
 		user->authToken = strdup (valueStr);
+	} else if (strcmp ("listenerId", key) == 0) {
+		user->listenerId = strdup (valueStr);
 	}
 }
 
@@ -301,14 +303,12 @@ static void PianoXmlParsePlaylistCb (const char *key, const ezxml_t value,
 		}
 	} else if (strcmp ("artistMusicId", key) == 0) {
 		song->artistMusicId = strdup (valueStr);
- 	} else if (strcmp ("testStrategy", key) == 0) {
-		song->testStrategy = atoi (valueStr);
-	} else if (strcmp ("songType", key) == 0) {
-		song->songType = atoi (valueStr);
 	} else if (strcmp ("feedbackId", key) == 0) {
 		song->feedbackId = strdup (valueStr);
 	} else if (strcmp ("songDetailURL", key) == 0) {
 		song->detailUrl = strdup (valueStr);
+	} else if (strcmp ("trackToken", key) == 0) {
+		song->trackToken = strdup (valueStr);
 	}
 }
 
@@ -831,6 +831,7 @@ struct PianoXmlParseSeedBag {
 	char *seedId;
 	PianoSong_t *song;
 	PianoArtist_t *artist;
+	PianoStation_t *station;
 };
 
 /*	parse seed struct
@@ -839,7 +840,11 @@ static void PianoXmlParseSeedCb (const char *key, const ezxml_t value,
 		void *data) {
 	struct PianoXmlParseSeedBag *bag = data;
 
+	assert (bag != NULL);
+
 	if (strcmp ("song", key) == 0) {
+		assert (bag->song == NULL);
+
 		if ((bag->song = calloc (1, sizeof (*bag->song))) == NULL) {
 			return;
 		}
@@ -847,12 +852,24 @@ static void PianoXmlParseSeedCb (const char *key, const ezxml_t value,
 		PianoXmlStructParser (ezxml_child (value, "struct"),
 				PianoXmlParsePlaylistCb, bag->song);
 	} else if (strcmp ("artist", key) == 0) {
+		assert (bag->artist == NULL);
+
 		if ((bag->artist = calloc (1, sizeof (*bag->artist))) == NULL) {
 			return;
 		}
 
 		PianoXmlStructParser (ezxml_child (value, "struct"),
 				PianoXmlParseSearchArtistCb, bag->artist);
+	} else if (strcmp ("nonGenomeStation", key) == 0) {
+		/* genre stations are "non genome" station seeds */
+		assert (bag->station == NULL);
+
+		if ((bag->station = calloc (1, sizeof (*bag->station))) == NULL) {
+			return;
+		}
+
+		PianoXmlStructParser (ezxml_child (value, "struct"),
+				PianoXmlParseStationsCb, bag->station);
 	} else if (strcmp ("seedId", key) == 0) {
 		char *valueStr = PianoXmlGetNodeText (value);
 		bag->seedId = strdup (valueStr);
@@ -875,10 +892,15 @@ static void PianoXmlParseGetStationInfoCb (const char *key, const ezxml_t value,
 			PianoXmlStructParser (ezxml_child (seedNode, "struct"),
 					PianoXmlParseSeedCb, &bag);
 
-			/* FIXME: use if-clause */
-			assert (bag.seedId != NULL);
-			assert (bag.song != NULL || bag.artist != NULL);
+			assert (bag.song != NULL || bag.artist != NULL ||
+					bag.station != NULL);
 
+			if (bag.seedId == NULL) {
+				/* seeds without id are useless */
+				continue;
+			}
+
+			/* FIXME: copy&waste */
 			if (bag.song != NULL) {
 				bag.song->seedId = bag.seedId;
 
@@ -903,6 +925,18 @@ static void PianoXmlParseGetStationInfoCb (const char *key, const ezxml_t value,
 					}
 					curSong->next = bag.artist;
 				}
+			} else if (bag.station != NULL) {
+				bag.station->seedId = bag.seedId;
+
+				if (info->stationSeeds == NULL) {
+					info->stationSeeds = bag.station;
+				} else {
+					PianoStation_t *curStation = info->stationSeeds;
+					while (curStation->next != NULL) {
+						curStation = curStation->next;
+					}
+					curStation->next = bag.station;
+				}
 			} else {
 				free (bag.seedId);
 			}
