@@ -136,7 +136,7 @@ static void BarMainGetInitialStation (BarApp_t *app) {
 	}
 	/* no autostart? ask the user */
 	if (app->curStation == NULL) {
-		app->curStation = BarUiSelectStation (app, "Select station: ", NULL);
+		app->curStation = BarUiSelectStation (app, app->ph.stations, "Select station: ", NULL);
 	}
 	if (app->curStation != NULL) {
 		BarUiPrintStation (&app->settings, app->curStation);
@@ -192,7 +192,7 @@ static void BarMainStartPlayback (BarApp_t *app, pthread_t *playerThread) {
 		/* setup player */
 		memset (&app->player, 0, sizeof (app->player));
 
-		WaitressInit (&app->player.waith);
+		WaitressInit (&app->player.waith, NULL);
 		WaitressSetUrl (&app->player.waith, app->playlist->audioUrl);
 
 		/* set up global proxy, player is NULLed on songfinish */
@@ -328,6 +328,7 @@ int main (int argc, char **argv) {
 	static BarApp_t app;
 	/* terminal attributes _before_ we started messing around with ~ECHO */
 	struct termios termOrig;
+	WaitressReturn_t wRet;
 
 	memset (&app, 0, sizeof (app));
 
@@ -338,12 +339,9 @@ int main (int argc, char **argv) {
 
 	/* init some things */
 	ao_initialize ();
+	gnutls_global_init ();
 	PianoInit (&app.ph);
 
-	WaitressInit (&app.waith);
-	app.waith.url.host = strdup (PIANO_RPC_HOST);
-	app.waith.url.port = strdup (PIANO_RPC_PORT);
-
 	BarSettingsInit (&app.settings);
 	BarSettingsRead (&app.settings);
 
@@ -357,6 +355,20 @@ int main (int argc, char **argv) {
 				app.settings.keys[BAR_KS_HELP]);
 	}
 
+	if ((wRet = WaitressInit (&app.waith, app.settings.tlsCaPath)) != WAITRESS_RET_OK) {
+		if (wRet == WAITRESS_RET_TLS_TRUSTFILE_ERR) {
+			BarUiMsg (&app.settings, MSG_ERR, "Can't load root certificates. "
+					"Please check the tls_ca_path setting in your config file.\n");
+		} else {
+			BarUiMsg (&app.settings, MSG_ERR, "Can't initialize HTTP library: "
+					"%s\n", WaitressErrorToStr (wRet));
+		}
+		goto die;
+	}
+
+	app.waith.url.host = strdup (PIANO_RPC_HOST);
+	app.waith.url.tls = true;
+
 	/* init fds */
 	FD_ZERO(&app.input.set);
 	app.input.fds[0] = STDIN_FILENO;
@@ -376,6 +388,7 @@ int main (int argc, char **argv) {
 
 	BarMainLoop (&app);
 
+die:
 	if (app.input.fds[1] != -1) {
 		close (app.input.fds[1]);
 	}
@@ -383,7 +396,9 @@ int main (int argc, char **argv) {
 	PianoDestroy (&app.ph);
 	PianoDestroyPlaylist (app.songHistory);
 	PianoDestroyPlaylist (app.playlist);
+	WaitressFree (&app.waith);
 	ao_shutdown();
+	gnutls_global_deinit ();
 	BarSettingsDestroy (&app.settings);
 
 	/* restore terminal attributes, zsh doesn't need this, bash does... */
